@@ -30,6 +30,11 @@
 #include <locale.h>
 #endif
 
+#ifdef HAVE_STRING_H
+#include <string.h>
+#endif
+
+#include <gio/gio.h>
 #include <exo/exo.h>
 #include <gtk/gtk.h>
 #include <libxfce4util/libxfce4util.h>
@@ -49,6 +54,8 @@ G_DEFINE_TYPE (ClipmanActions, clipman_actions, G_TYPE_OBJECT)
 
 struct _ClipmanActionsPrivate
 {
+  GFile                *file;
+  GFileMonitor         *file_monitor;
   GSList               *entries;
   GtkWidget            *menu;
 };
@@ -61,6 +68,7 @@ static void             clipman_actions_finalize            (GObject *object);
  * Misc functions declarations
  */
 
+static void            _clipman_actions_free_list           (ClipmanActions *actions);
 static gint           __clipman_actions_entry_compare       (gpointer a,
                                                              gpointer b);
 static gint           __clipman_actions_entry_compare_name  (gpointer a,
@@ -81,6 +89,11 @@ static void           __foreach_command_write_xml           (gpointer key,
 
 static void             cb_entry_activated                  (GtkMenuItem *mi,
                                                              gpointer user_data);
+static void             cb_file_changed                     (ClipmanActions *actions,
+                                                             GFile *file,
+                                                             GFile *other_file,
+                                                             GFileMonitorEvent event_type);
+static gboolean         timeout_file_changed                (ClipmanActions *actions);
 
 /*
  * XML Parser declarations
@@ -118,6 +131,7 @@ typedef enum
   ACTION,
   ACTION_NAME,
   REGEX,
+  GROUP,
   COMMANDS,
   COMMAND,
   COMMAND_NAME,
@@ -136,6 +150,7 @@ struct _EntryParser
 
   gchar *action_name;
   gchar *regex;
+  gint group;
   gchar *command_name;
   gchar *command;
 };
@@ -211,6 +226,8 @@ start_element_handler (GMarkupParseContext *context,
         }
       else if (!g_ascii_strcasecmp (element_name, "regex"))
         parser->state = REGEX;
+      else if (!g_ascii_strcasecmp (element_name, "group"))
+        parser->state = GROUP;
       else if (!g_ascii_strcasecmp (element_name, "commands"))
         parser->state = COMMANDS;
       else if (!g_ascii_strcasecmp (element_name, "exec"))
@@ -237,22 +254,29 @@ end_element_handler (GMarkupParseContext *context,
       g_free (parser->regex);
       parser->action_name = NULL;
       parser->regex = NULL;
+      parser->group = 0;
 
       parser->state = ACTIONS;
       break;
 
     case ACTION_NAME:
     case REGEX:
+    case GROUP:
     case COMMANDS:
       parser->state = ACTION;
       break;
 
     case COMMAND:
       if (parser->action_name == NULL || parser->regex == NULL)
-        g_warning ("Closing a command but no action name nor regex set");
+        {
+          g_warning ("Closing a command but no action name nor regex set");
+        }
       else
-        clipman_actions_add (parser->actions, parser->action_name, parser->regex,
-                             parser->command_name, parser->command);
+        {
+          clipman_actions_add (parser->actions, parser->action_name, parser->regex,
+                               parser->command_name, parser->command);
+          clipman_actions_set_group (parser->actions, parser->action_name, parser->group);
+        }
 
       g_free (parser->command_name);
       g_free (parser->command);
@@ -295,6 +319,9 @@ text_handler (GMarkupParseContext *context,
       parser->regex = g_strdup (text);
       break;
 
+    case GROUP:
+      parser->group = (gint)g_strtod (text, NULL);
+
     case COMMAND_NAME:
       if (parser->name_use)
         {
@@ -343,10 +370,43 @@ cb_entry_activated (GtkMenuItem *mi,
   g_free (real_command);
 }
 
+static void
+cb_file_changed (ClipmanActions *actions,
+                 GFile *file,
+                 GFile *other_file,
+                 GFileMonitorEvent event_type)
+{
+  static guint timeout = 0;
+  if (event_type == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT)
+    {
+      if (timeout > 0)
+        g_source_remove (timeout);
+      timeout = g_timeout_add_seconds (1, (GSourceFunc)timeout_file_changed, actions);
+    }
+}
+
+static gboolean
+timeout_file_changed (ClipmanActions *actions)
+{
+  _clipman_actions_free_list (actions);
+  clipman_actions_load (actions);
+  return FALSE;
+}
+
 /*
  * Misc functions
  */
 
+static void
+_clipman_actions_free_list (ClipmanActions *actions)
+{
+  GSList *l;
+  for (l = actions->priv->entries; l != NULL; l = l->next)
+    __clipman_actions_entry_free (l->data);
+  g_slist_free (actions->priv->entries);
+  actions->priv->entries = NULL;
+}
+
 static gint
 __clipman_actions_entry_compare (gpointer a,
                                  gpointer b)
@@ -437,6 +497,9 @@ __foreach_command_write_xml (gpointer key,
  * The command can contain the parameter '%s' that will be replaced by the
  * matching regex text.
  *
+ * The action is created with the default group 0. To change it use the
+ * function clipman_actions_set_group().
+ *
  * Returns: FALSE if the regex was invalid
  */
 gboolean
@@ -469,6 +532,7 @@ clipman_actions_add (ClipmanActions *actions,
       entry = g_slice_new0 (ClipmanActionsEntry);
       entry->action_name = g_strdup (action_name);
       entry->regex = _regex;
+      entry->group = 0;
       entry->commands = g_hash_table_new_full ((GHashFunc)g_str_hash, (GEqualFunc)g_str_equal,
                                                (GDestroyNotify)g_free, (GDestroyNotify)g_free);
       g_hash_table_insert (entry->commands, g_strdup (command_name), g_strdup (command));
@@ -564,6 +628,33 @@ clipman_actions_remove_command (ClipmanActions *actions,
   return found;
 }
 
+/**
+ * clipman_actions_set_group:
+ * @actions:            a #ClipmanActions
+ * @action_name:        the human readable name for the regex
+ * @group:              the group identifier
+ *
+ * Changes the group of @action_name to @group.
+ */
+void
+clipman_actions_set_group (ClipmanActions *actions,
+                           const gchar *action_name,
+                           gint group)
+{
+  ClipmanActionsEntry *entry;
+  GSList *l;
+
+  l = g_slist_find_custom (actions->priv->entries, action_name, (GCompareFunc)__clipman_actions_entry_compare_name);
+  if (l == NULL)
+    {
+      g_warning ("No corresponding entry `%s'", action_name);
+      return;
+    }
+
+  entry = l->data;
+  entry->group = group;
+}
+
 /**
  * clipman_actions_get_entries:
  * @actions:    a #ClipmanActions
@@ -579,6 +670,7 @@ clipman_actions_get_entries (ClipmanActions *actions)
 /**
  * clipman_actions_match:
  * @actions:    a #ClipmanActions
+ * @group:      the group identifier
  * @text:       the text to match against the existing regex's
  *
  * Searches a regex match for @text and returns a newly allocated #GSList which
@@ -586,11 +678,13 @@ clipman_actions_get_entries (ClipmanActions *actions)
  * #ClipmanActionsEntry for each matched action.
  * Note that the data inside the list is owned by #ClipmanActions and must not
  * be modified.
+ * The @group identifier can be -1 to get a match from all groups.
  *
  * Returns: a newly allocated #GSList
  */
 GSList *
 clipman_actions_match (ClipmanActions *actions,
+                       gint group,
                        const gchar *text)
 {
   ClipmanActionsEntry *entry;
@@ -600,8 +694,11 @@ clipman_actions_match (ClipmanActions *actions,
   for (l = actions->priv->entries; l != NULL; l = l->next)
     {
       entry = l->data;
-      if (g_regex_match (entry->regex, text, 0, NULL))
-        entries = g_slist_prepend (entries, entry);
+      if (group == -1 || group == entry->group)
+        {
+          if (g_regex_match (entry->regex, text, 0, NULL))
+            entries = g_slist_prepend (entries, entry);
+        }
     }
 
   return entries;
@@ -610,6 +707,7 @@ clipman_actions_match (ClipmanActions *actions,
 /**
  * clipman_actions_match_with_menu:
  * @actions:    a #ClipmanActions
+ * @group:      the group identifier
  * @text:       the text to match against the existing regex's
  *
  * Builds and displays a menu with matching actions for @text.
@@ -618,13 +716,14 @@ clipman_actions_match (ClipmanActions *actions,
  */
 void
 clipman_actions_match_with_menu (ClipmanActions *actions,
+                                 gint group,
                                  const gchar *text)
 {
   ClipmanActionsEntry *entry;
   GtkWidget *mi;
   GSList *l, *entries;
 
-  entries = clipman_actions_match (actions, text);
+  entries = clipman_actions_match (actions, group, text);
 
   if (entries == NULL)
     return;
@@ -692,40 +791,38 @@ clipman_actions_load (ClipmanActions *actions)
 {
   gchar *filename;
   gchar *data;
-  gsize size;
+  gssize size;
   gboolean load;
   GMarkupParseContext *context;
   EntryParser *parser;
 
-  filename = g_strdup_printf ("%s/xfce4/panel/xfce4-clipman-actions.xml", g_get_user_config_dir ());
-  load = g_file_get_contents (filename, &data, &size, NULL);
+  load = g_file_load_contents (actions->priv->file, NULL, &data, &size, NULL, NULL);
 
   if (!load)
     {
-      g_free (filename);
       filename = g_strdup (SYSCONFDIR"/xdg/xfce4/panel/xfce4-clipman-actions.xml");
       load = g_file_get_contents (filename, &data, &size, NULL);
+      g_free (filename);
     }
 
   if (!load)
-    g_warning ("Unable to load actions from an XML file");
-
-  DBG ("Load actions from file %s", filename);
-
-  if (load)
     {
-      parser = g_slice_new0 (EntryParser);
-      parser->actions = actions;
-      parser->locale = setlocale (LC_MESSAGES, NULL);
-      context = g_markup_parse_context_new (&markup_parser, 0, parser, NULL);
-      g_markup_parse_context_parse (context, data, (gssize)size, NULL);
-      if (!g_markup_parse_context_end_parse (context, NULL))
-        g_warning ("Error parsing the XML file");
-      g_markup_parse_context_free (context);
-      g_slice_free (EntryParser, parser);
+      g_warning ("Unable to load actions from an XML file");
+      return;
     }
 
-  g_free (filename);
+  DBG ("Load actions from file");
+
+  parser = g_slice_new0 (EntryParser);
+  parser->actions = actions;
+  parser->locale = setlocale (LC_MESSAGES, NULL);
+  context = g_markup_parse_context_new (&markup_parser, 0, parser, NULL);
+  g_markup_parse_context_parse (context, data, size, NULL);
+  if (!g_markup_parse_context_end_parse (context, NULL))
+    g_warning ("Error parsing the XML file");
+  g_markup_parse_context_free (context);
+  g_slice_free (EntryParser, parser);
+
   g_free (data);
 }
 
@@ -738,7 +835,6 @@ void
 clipman_actions_save (ClipmanActions *actions)
 {
   ClipmanActionsEntry *entry;
-  gchar *filename;
   gchar *data;
   GString *output;
   gchar *tmp;
@@ -762,6 +858,8 @@ clipman_actions_save (ClipmanActions *actions)
       g_string_append_printf (output, "\t\t<regex>%s</regex>\n", tmp);
       g_free (tmp);
 
+      g_string_append_printf (output, "\t\t<group>%d</group>\n", entry->group);
+
       g_string_append (output, "\t\t<commands>\n");
 
 #if GLIB_CHECK_VERSION (2,16,0)
@@ -794,13 +892,12 @@ clipman_actions_save (ClipmanActions *actions)
   g_string_append (output, "</actions>");
 
   /* And now write output to the xml file */
-  filename = g_strdup_printf ("%s/xfce4/panel/xfce4-clipman-actions.xml", g_get_user_config_dir ());
-  DBG ("Save actions to file %s", filename);
+  DBG ("Save actions to file");
   data = g_string_free (output, FALSE);
-  if (!g_file_set_contents (filename, data, -1, NULL))
-    g_warning ("Unable to write the actions to the XML file %s", filename);
+  if (!g_file_replace_contents (actions->priv->file, data, strlen (data), NULL, FALSE,
+                                G_FILE_CREATE_NONE, NULL, NULL, NULL))
+    g_warning ("Unable to write the actions to the XML file");
 
-  g_free (filename);
   g_free (data);
 }
 
@@ -844,19 +941,29 @@ clipman_actions_class_init (ClipmanActionsClass *klass)
 static void
 clipman_actions_init (ClipmanActions *actions)
 {
+  gchar *filename;
+
   actions->priv = GET_PRIVATE (actions);
 
+  /* Actions file */
+  filename = g_strdup_printf ("%s/xfce4/panel/xfce4-clipman-actions.xml", g_get_user_config_dir ());
+  actions->priv->file = g_file_new_for_path (filename);
+  g_free (filename);
+
+  /* Load initial actions */
   clipman_actions_load (actions);
+
+  /* Listen on xml file changes */
+  actions->priv->file_monitor = g_file_monitor_file (actions->priv->file, G_FILE_MONITOR_NONE, NULL, NULL);
+  g_signal_connect_swapped (actions->priv->file_monitor, "changed", G_CALLBACK (cb_file_changed), actions);
 }
 
 static void
 clipman_actions_finalize (GObject *object)
 {
   ClipmanActions *actions = CLIPMAN_ACTIONS (object);
-  GSList *l;
-
-  for (l = actions->priv->entries; l != NULL; l = l->next)
-    __clipman_actions_entry_free (l->data);
-  g_slist_free (actions->priv->entries);
+  _clipman_actions_free_list (actions);
+  g_object_unref (actions->priv->file_monitor);
+  g_object_unref (actions->priv->file);
 }
 
