@@ -23,13 +23,14 @@
  */
 
 /* History */
-#define DEFAULT_MAX_TEXTS_IN_HISTORY            10
-#define DEFAULT_MAX_IMAGES_IN_HISTORY           1
-#define DEFAULT_SAVE_ON_QUIT                    TRUE
+#define DEFAULT_MAX_TEXTS_IN_HISTORY                    10
+#define DEFAULT_MAX_IMAGES_IN_HISTORY                   1
+#define DEFAULT_SAVE_ON_QUIT                            TRUE
 
 /* Collector */
-#define DEFAULT_ADD_PRIMARY_CLIPBOARD           FALSE
-#define DEFAULT_ENABLE_ACTIONS                  FALSE
+#define DEFAULT_ADD_PRIMARY_CLIPBOARD                   FALSE
+#define DEFAULT_HISTORY_IGNORE_PRIMARY_CLIPBOARD        TRUE
+#define DEFAULT_ENABLE_ACTIONS                          FALSE
 
 /*
  * Selection for the popup command
@@ -38,3 +39,9 @@
 #define XFCE_CLIPMAN_SELECTION    "XFCE_CLIPMAN_SELECTION"
 #define XFCE_CLIPMAN_MESSAGE      "MENU"
 
+/*
+ * Action Groups
+ */
+
+#define ACTION_GROUP_SELECTION  0
+#define ACTION_GROUP_MANUAL     1
